# -*- coding: utf-8 -*-
import inspect


from .DomainUtils import DomainUtils


class PatternBurstBlockParser:
    def __init__(self, debug=False):
        self.debug = debug

        # List with all pattern burst block names
        self.patt_burst_names = []

        # dict key   is the pattern burst block name
        # dict value is a list with pattern blocks
        self.patt_burst_block2pattern_blocks = {}

        # dict key   is the pattern name
        # dict value is lark token object which holds the pattern name
        self.patt2obj = {}

        # dict key   is the macrodefs domain name
        # dict value is lark token object which holds the macrodefs domain name
        self.macro_domain2obj = {}

        # dict key   is the procedures domain name
        # dict value is lark token object which holds the procedure domain name
        self.proc_domain2obj = {}

        # dict key   is the pattern burst name
        # dict value is macrodefs domain name
        self.patt_burst2macro_domain = {}

        # dict key   is the pattern burst name
        # dict value is procedures domain name
        self.patt_burst2proc_domain = {}

        # dict key   is the pattern burst name
        # dict value is signal groups domain name
        self.patt_burst2sig_groups_domain = {}

        self.curr_patt_burst = ""

        # dict key   is the pattern name
        # dict value is a list with all signal groups domain defined for this pattern name
        self.patt2sig_group_domain = {}

        self.curr_sig_group_domain = DomainUtils.global_domain

        # dict key   is the pattern name
        # dict value is a list with all timing domains defined for this pattern name in the pattern exec block
        self.patt2timing_domain = {}

        # dict key   is the macro domain
        # dict value is a list with all timing domains defined for this macro domain in the pattern exec block
        self.macro_domain2timing_domain = {}

        # dict key   is the procedure domain
        # dict value is a list with all timing domains defined for this procedure domain in the pattern exec block
        self.proc_domain2timing_domain = {}

        self.is_patternburst_block_defined = False

    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')

    def b_pattern_burst__PATTERN_BURST_BLOCK_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if t.value in self.patt_burst_block2pattern_blocks:
            err_msg = f"Pattern burst block {t} already defined! Pattern burst block names must be unique."
            raise Exception(err_msg)
        else:
            self.patt_burst_block2pattern_blocks[t.value] = []
            self.curr_patt_burst = t.value

        global_domain = DomainUtils.global_domain
        self.last_proc_domain = global_domain
        self.patt_burst2macro_domain[self.curr_patt_burst] = global_domain
        self.patt_burst2proc_domain[self.curr_patt_burst] = global_domain
        self.patt_burst2sig_groups_domain[self.curr_patt_burst] = global_domain

    def b_pattern_burst__OPEN_PATTERN_BURST_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if self.is_signal_block_defined == False:
            err_msg = "Expected Signal block before any PatternBurst block"
            raise Exception(err_msg)
        if self.is_signalgroups_block_defined == False:
            err_msg = "Expected SignalGroups block before any PatternBurst block"
            raise Exception(err_msg)
        if self.is_timing_block_defined == False:
            err_msg = "Expected Timing block before any PatternBurst block"
            raise Exception(err_msg)
        self.is_patternburst_block_defined = True

        if self.curr_patt_burst in self.patt_burst_names:
            err_msg = f"Pattern burst name '{self.curr_patt_burst}' is already defined"
            raise Exception(err_msg)
        else:
            self.patt_burst_names.append(self.curr_patt_burst)

    def b_pattern_burst__PATT_OR_BURST_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        patt_burst_block = self.patt_burst_block2pattern_blocks[self.curr_patt_burst]
        # Collect pattern token to check later if it is defined at the end of the file
        self.patt2obj[t.value] = t
        
        if t.value not in self.patt_burst_names:
            patt = t.value
            patt_burst_block.append(patt)
            if patt in self.patt2sig_group_domain:
                sig_group_list = self.patt2sig_group_domain[t.value]
                if self.curr_sig_group_domain not in sig_group_list:
                    sig_group_list.append(self.curr_sig_group_domain)
            else:
                self.patt2sig_group_domain[patt] = [self.curr_sig_group_domain]
        else:
            err_msg = f"No support so far for pattern burst reference {t.value} in pattern burst block"
            raise Exception(err_msg)
        

    def b_pattern_burst__SIGNAL_GROUPS_DOMAIN(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.patt_burst2sig_groups_domain[self.curr_patt_burst] = t.value
        if t.value not in self.signal_groups_domains:
            err_msg = f"Signal Groups domain name {t.value} used in pattern burst {self.curr_patt_burst} is not defined!"
            raise Exception(err_msg)
        self.curr_sig_group_domain = t.value

    #            raise VisitError(err_msg, t, Exception(err_msg))

    def b_pattern_burst__MACROS_DOMAIN(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.patt_burst2macro_domain[self.curr_patt_burst] = t.value
        self.macro_domain2obj[t.value] = t

    def b_pattern_burst__PROCEDURES_DOMAIN(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.patt_burst2proc_domain[self.curr_patt_burst] = t.value
        self.proc_domain2obj[t.value] = t

    def b_pattern_burst__CLOSE_PATTERN_BURST_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

            print("Found pattern blocks in the PatternBurst block:")
            for key in self.patt_burst_block2pattern_blocks.keys():
                print(f" PatternBurst block {key} ")
                for value in self.patt_burst_block2pattern_blocks[key]:
                    print(f"  => Pattern/Burst block : {value} ")
        self.curr_sig_group_domain = DomainUtils.global_domain
        self.curr_patt_burst = ""



# -*- coding: utf-8 -*-

class TimeUtils:
    
    
    def bodmas(data, time_expr):
        
#        print(f"bodmas input {data}")

        # Search for brackets
        start_brackets = []
        end_brackets = []
        
        index = 0
        for item in data:
            if item == '(':
                start_brackets.append(index)
            elif item == ')':
                end_brackets.append(index)
            index += 1
            
        sb = len(start_brackets)
        eb = len(end_brackets)
        
        if (sb > 0 or eb > 0) and sb != eb:
            err_msg = f"No matching brаckets for time expression {time_expr}!"
            raise Exception(err_msg)
        elif sb > 0 and eb > 0:
            for i in range(sb):
                sbi = start_brackets[i]
                ebi = end_brackets[i]
                sub_data = []
                for di in range(sbi+1, ebi):
                    d = data[di]
                    sub_data.append(d)
                val = TimeUtils.bodmas(sub_data, time_expr)
                
#                print(f"bodmas after recursive () {val}")
                
#                print(f"bodmas before substitution () {data}")
                insert = sbi
                del data[sbi:ebi+1]
                data.insert(insert, str(val))
#                print(f"bodmas after substitution () {data}")

#        print(f"bodmas after () {data}")

        div = []
        
        index = 0
        for item in data:
            if item == '/':
                div.append(index)
            index += 1
        
        if len(div) > 0:
            for i in range(len(div)):
                
                left  = data[div[i]-1]
                right = data[div[i]+1]
                
                # Two cases:
                # Numeric / Frequency
                # Time / Numeric

                try:
                    left_value = float(left)
                    freq = TimeUtils.get_freq_Hz(right)
                    if freq != None:
                        result = left_value / freq * 1E15
                    else:
                        err_msg = f"Expecting frequency value for the right operand, but got '{right}' in time expression '{time_expr}'!"
                        raise Exception(err_msg)
                except:
                    try:
                        right_value = float(right)
                        left_value = TimeUtils.get_time_fsec(left)
                        result = left_value / right_value
                    except:
                        err_msg = f"Expecting either numeric value for the left or right operand, but got left '{left}' and right '{right}' in time expression '{time_expr}'!"
                        raise Exception(err_msg)
                
                insert = div[i]-1
                del data[div[i]-1:div[i]+2]
                data.insert(insert, str(int(result)) + "fs")
                
#        print(f"bodmas after / {data}")
                    
        mul = []
        
        index = 0
        for item in data:
            if item == '*':
                mul.append(index)
            index += 1
                
        if len(mul) > 0:
            for i in range(len(mul)):
                
                left  = data[mul[i]-1]
                right = data[mul[i]+1]
                
                left_value = TimeUtils.get_time_fsec(left)
                right_value = TimeUtils.get_time_fsec(right)
                
                is_right_value_time = True
                
                if left_value == None:
                    try:
                        float(left)
                    except:
                        err_msg = f"Expecting time or numeric value for the left operand, but got '{left}' in time expression '{time_expr}'!"
                        raise Exception(err_msg)
                elif right_value == None:
                    try:
                        right_value = float(right)
                        is_right_value_time = False
                    except:
                        err_msg = f"Expecting time or numeric value for the right operand, but got '{right}' in time expression '{time_expr}'!"
                        raise Exception(err_msg)
                
                if is_right_value_time:
                    result = left_value * right_value / 1E6
                else:
                    result = left_value * right_value

                insert = mul[i]-1
                del data[mul[i]-1:mul[i]+2]
                data.insert(insert, str(int(result)) + "fs")
                
#        print(f"bodmas after * {data}")

        add = []
        
        index = 0
        for item in data:
            if item == '+':
                add.append(index)
            index += 1

        if len(add) > 0:
            for i in range(len(add)):
                
                left  = data[add[i]-1]
                right = data[add[i]+1]
                
                left_value = TimeUtils.get_time_fsec(left)
                right_value = TimeUtils.get_time_fsec(right)
                
                if left_value == None:
                    err_msg = f"Expecting time value for the left operand, but got '{left}' in time expression '{time_expr}'!"
                    raise Exception(err_msg)
                elif right_value == None:
                    err_msg = f"Expecting time value for the right operand, but got '{right}' in time expression '{time_expr}'!"
                    raise Exception(err_msg)
                
                result = left_value + right_value
                insert = add[i]-1
                del data[add[i]-1:add[i]+1]
                data.insert(insert, str(int(result)) + "fs")

#        print(f"bodmas after - {data}")

        sub = []
        
        index = 0
        for item in data:
            if item == '-':
                sub.append(index)
            index += 1

        if len(sub) > 0:
            for i in range(len(sub)):
                left  = data[sub[i]-1]
                right = data[sub[i]+1]
                
                left_value = TimeUtils.get_time_fsec(left)
                right_value = TimeUtils.get_time_fsec(right)
                
                if left_value == None:
                    err_msg = f"Expecting time value for the left operand, but got '{left}' in time expression '{time_expr}'!"
                    raise Exception(err_msg)
                elif right_value == None:
                    err_msg = f"Expecting time value for the right operand, but got '{right}' in time expression '{time_expr}'!"
                    raise Exception(err_msg)
                
                result = left_value - right_value
                insert = sub[i]-1
                del data[sub[i]-1:sub[i]+1]
                data.insert(insert, str(int(result)) + "fs")

#        print(f"bodmas after + {data}")
#        print(f"bodmas return {data}")
        
        if data[0].endswith('fs'):
            return data[0]
        else:
            fsec = TimeUtils.get_time_fsec(data[0])
            return str(int(fsec)) + "fs"

    def get_freq_Hz(expr):
        
        v = expr.strip()
        if v.endswith('Hz'):
            
            if v.endswith('aHz'):
                v_str = v[0:len(v)-3]
                value = float(v_str)
                return value/1000000000000000000.0
            elif v.endswith('fHz'):
                v_str = v[0:len(v)-3]
                value = float(v_str)
                return value/1000000000000000.0
            elif v.endswith('pHz'):
                v_str = v[0:len(v)-3]
                value = float(v_str)
                return value/1000000000000.0
            elif v.endswith('nHz'):
                v_str = v[0:len(v)-3]
                value = float(v_str)
                return value/1000000000.0
            elif v.endswith('uHz'):
                v_str = v[0:len(v)-3]
                value = float(v_str)
                return value/1000000.0
            elif v.endswith('mHz'):
                v_str = v[0:len(v)-3]
                value = float(v_str)
                return value/1000.0

            elif v.endswith('kHz'):
                v_str = v[0:len(v)-3]
                value = float(v_str)
                return value*1000.0
            elif v.endswith('MHz'):
                v_str = v[0:len(v)-3]
                value = float(v_str)
                return value*1000000.0
            elif v.endswith('GHz'):
                v_str = v[0:len(v)-3]
                value = float(v_str)
                return value*1000000000.0
            elif v.endswith('THz'):
                v_str = v[0:len(v)-3]
                value = float(v_str)
                return value*1000000000000.0
            elif v.endswith('PHz'):
                v_str = v[0:len(v)-3]
                value = float(v_str)
                return value*1000000000000000.0
            elif v.endswith('EHz'):
                v_str = v[0:len(v)-3]
                value = float(v_str)
                return value*1000000000000000000.0
            else:
                # Hz
                v_str = v[0:len(v)-1]
                value = float(v_str)
                return value
        else:
            return None



    def get_time_fsec(expr):
        
        v = expr.strip()
        if v[0] == "'":
            v = v[1:len(v)]
        
        if v[-1] == "'":
            v = v[0:-1]
            
        if v.endswith('s'):
            
            if v.endswith('as'):
                v_str = v[0:len(v)-2]
                value = float(v_str)
                return value/1000
            elif v.endswith('fs'):
                v_str = v[0:len(v)-2]
                value = float(v_str)
                return value
            elif v.endswith('ps'):
                v_str = v[0:len(v)-2]
                value = float(v_str)
                return value*1000
            elif v.endswith('ns'):
                v_str = v[0:len(v)-2]
                value = float(v_str)
                return value*1000000
            elif v.endswith('us'):
                v_str = v[0:len(v)-2]
                value = float(v_str)
                return value*1000000000
            elif v.endswith('ms'):
                v_str = v[0:len(v)-2]
                value = float(v_str)
                return value*1000000000000

            elif v.endswith('ks'):
                v_str = v[0:len(v)-2]
                value = float(v_str)
                return value*1000000000000000000
            elif v.endswith('Ms'):
                v_str = v[0:len(v)-2]
                value = float(v_str)
                return value*1000000000000000000000
            elif v.endswith('Gs'):
                v_str = v[0:len(v)-2]
                value = float(v_str)
                return value*1000000000000000000000000
            elif v.endswith('Ts'):
                v_str = v[0:len(v)-2]
                value = float(v_str)
                return value*1000000000000000000000000000
            elif v.endswith('Ps'):
                v_str = v[0:len(v)-2]
                value = float(v_str)
                return value*1000000000000000000000000000000
            elif v.endswith('Es'):
                v_str = v[0:len(v)-2]
                value = float(v_str)
                return value*1000000000000000000000000000000000
            else:
                # seconds
                v_str = v[0:len(v)-1]
                value = float(v_str)
                return value*1000000000000000
        else:
            return None



# -*- coding: utf-8 -*-
import inspect
from .DomainUtils import DomainUtils


class SpecBlockParser:
    def __init__(self, debug=False):
        self.debug = debug

        self.curr_spec = DomainUtils.global_domain
        self.curr_var_name = None
        self.curr_var_type = None
        self.curr_category = 'NONE'

        # key is category::variable      
        # If category is not used, the value will be "NONE"
        # value is the variable value        
        self.var_min_value = {}
        self.var_typ_value = {}
        self.var_max_value = {}

        self.variables = []

    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')

    def b_spec__SPEC_DOMAIN_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.curr_spec = t.value

    def b_spec__OPEN_SPEC_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

    def b_spec__cat_name(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.curr_category = str(t[0])

    def b_spec__var_name(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.curr_var_name = str(t[0])
        
        if self.curr_var_name not in self.variables:
            self.variables.append(self.curr_var_name)

    def b_spec__var_typ_value(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
            
        key = self.curr_spec + "::" + self.curr_category + "::" + self.curr_var_name
        self.var_typ_value[key] = str(t[0])

    def b_spec__var_type(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        
        self.curr_var_type = str(t[0])

    def b_spec__var_value(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
            
        key = self.curr_category + "::" + self.curr_var_name
        value = str(t[0]).replace('\'','')
            
        if self.curr_var_type == "Min":
            self.var_min_value[key] = value
        elif self.curr_var_type == "Typ":
            self.var_typ_value[key] = value
        elif self.curr_var_type == "Max":
            self.var_max_value[key] = value
                    
            
    def b_spec__close_category_block(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.curr_category = 'None'

    def b_spec__CLOSE_SPEC_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)



# -*- coding: utf-8 -*-


class DomainUtils:

    global_domain = "*global*"
    domain_separator = "::"

    def __init__(self, debug=False):
        pass

    def get_full_name(domain_name, user_defined_name):
        return domain_name + DomainUtils.domain_separator + user_defined_name

    def get_name(user_defined_name):
        pos = user_defined_name.find(DomainUtils.domain_separator)
        if pos > 0:
            return user_defined_name[pos + 2 :]
        else:
            return user_defined_name

    def get_domain(user_defined_name, is_user_friendly=False):
        pos = user_defined_name.find(DomainUtils.domain_separator)
        if pos > 0:
            ret_value = user_defined_name[:pos]
            if is_user_friendly and ret_value == DomainUtils.global_domain:
                return "unnamed"
            else:
                return ret_value
        else:
            if is_user_friendly and user_defined_name == DomainUtils.global_domain:
                return "unnamed"
            else:
                return user_defined_name



# -*- coding: utf-8 -*-
import inspect

from lark import Transformer

from . import STILBlockParser
from . import HeaderBlockParser
from . import SignalsBlockParser
from . import SignalGroupsBlockParser
from . import TimingBlockParser
from . import ScanStructBlockParser
from . import SelectorBlockParser
from . import SpecBlockParser
from . import MacroDefsBlockParser
from . import ProceduresBlockParser
from . import PatternBurstBlockParser
from . import PatternBlockParser
from . import PatternExecBlockParser


class STILLark(
    Transformer,
    STILBlockParser,
    HeaderBlockParser,
    SignalsBlockParser,
    SignalGroupsBlockParser,
    TimingBlockParser,
    ScanStructBlockParser,
    SpecBlockParser,
    SelectorBlockParser,
    MacroDefsBlockParser,
    ProceduresBlockParser,
    PatternBurstBlockParser,
    PatternExecBlockParser,
    PatternBlockParser,
):
    def __init__(self, stil_file, debug=False):

        """
        Parameters
        ----------
        stil : str
            stil argument can be either location of STIL file or
            the STIL file content as string.
        debug : TYPE, optional
            Debug flag. The default is False.

        Returns
        -------
        Nothing

        """
        self.debug = debug
        self.stil_file = stil_file
        self.total_va = 0
        self.user_defined_keywords = set()

        STILBlockParser.__init__(self, debug)
        HeaderBlockParser.__init__(self, debug)
        SignalsBlockParser.__init__(self, debug)
        SignalGroupsBlockParser.__init__(self, debug)
        TimingBlockParser.__init__(self, debug)
        ScanStructBlockParser.__init__(self, debug)
        SelectorBlockParser.__init__(self, debug)
        SpecBlockParser.__init__(self, debug)
        MacroDefsBlockParser.__init__(self, debug)
        ProceduresBlockParser.__init__(self, debug)
        PatternBurstBlockParser.__init__(self, debug)
        PatternExecBlockParser.__init__(self, debug)
        PatternBlockParser.__init__(self, debug)

    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')

    def USER_KEYWORDS(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.user_defined_keywords.add(t.value)
            
    def eof(self):
        # end of the file
        PatternBlockParser.eof(self)



# -*- coding: utf-8 -*-
import inspect


class SignalsBlockParser:
    def __init__(self, debug=False):
        self.debug = debug
        # dict key   is the signal name
        # dict value is the signal type
        self.sig2type = {}
        # Current processing signal
        self.curr_signal = ""

        self.is_signal_block_defined = False

        # dict key   is the signal name
        # dict value is the default drive state of the signal if not used
        self.sig2def_state = {}

        # List of scan in signals defined in the signals block
        self.scanin_sigs = []

        # List of scan out signals defined in the signals block
        self.scanout_sigs = []

    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')

    def b_signals__OPEN_SIGNAL_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if self.is_signal_block_defined:
            err_msg = "Duplication of Signals block found. Only one is allowed !"
            raise Exception(err_msg)

        self.is_signal_block_defined = True

    def b_signals__SIGNAL_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        if t.value in self.sig2type:
            err_msg = f"Signal {t} already defined! Signal names in Signals block must be unique."
            raise Exception(err_msg)
        else:
            self.sig2type[t] = ""
            self.curr_signal = t.value

    def b_signals__SIGNAL_TYPE(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.sig2type[self.curr_signal] = t.value
        
        
    def b_signals__KEYWORD_SCANIN(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        
        if self.curr_signal not in self.scanin_sigs:
            self.scanin_sigs.append(self.curr_signal)
        
    def b_signals__KEYWORD_SCANOUT(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if self.curr_signal not in self.scanout_sigs:
            self.scanout_sigs.append(self.curr_signal)

    def b_signals__SIGNAL_END_STMT(self, t):
        self.curr_signal = ""
        
    def b_signals__CLOSE_SIGNAL_ATTR_BLOCK(self, t):
        self.curr_signal = ""        

    def b_signals__SIG_DEF_STATE_UP(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.sig2def_state[self.curr_signal] = "U"

    def b_signals__SIG_DEF_STATE_DOWN(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.sig2def_state[self.curr_signal] = "D"

    def b_signals__SIG_DEF_STATE_OFF(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.sig2def_state[self.curr_signal] = "Z"

    def b_signals__CLOSE_SIGNAL_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

            print("Found signals in the Signal block:")
            for signal_name in self.sig2type:
                print(f" Signal {signal_name} has type {self.sig2type[signal_name]}")



# -*- coding: utf-8 -*-
import inspect


class STILBlockParser:
    def __init__(self, debug=False):
        self.debug = debug
        self.extensions = {}
        self.last_extension = ""

    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')

    def b_stil__STIL_VERSION(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        if t != "1.0":
            err_msg = "Not supported STIL file format : "
            err_msg += "expected 1.0, but the value is " + t
            raise Exception(err_msg)

    def b_stil__OPEN_STIL_EXTENSION(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

    def b_stil__STIL_EXTENSION_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.last_extension = t

    def b_stil__STIL_EXTENSION_YEAR(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        if self.last_extension == "Design" and t != "2005":
            err_msg = "Not supported STIL extension year : "
            err_msg += "expected 2005, but the value is " + t
            raise Exception(err_msg)
        elif self.last_extension == "DCLevels" and t != "2002":
            err_msg = "Not supported STIL extension year : "
            err_msg += "expected 2002, but the value is " + t
            raise Exception(err_msg)
        else:
            self.extensions[self.last_extension] = t

    def b_stil__CLOSE_STIL_EXTENSION(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)



# -*- coding: utf-8 -*-
import os
import gzip

def check_for_compression(file):

    if os.path.exists(file):
        #Check for gz compression
        with open(file, 'rb') as zf:
            if zf.read(2) == b'\x1f\x8b':
                return True
        #Check for zip compression
        with open(file, 'rb') as zf:
            if zf.read(4) == b'\x50\x4B\x03\x04':
                return True
    return False

def get_uncompressed_data(file):

    if os.path.exists(file):
        is_gz = False
        is_zip = False
        
        #Check for gz compression
        with open(file, 'rb') as zf:
            if zf.read(2) == b'\x1f\x8b':
                is_gz = True
                
        if is_gz == False:
            #Check for zip compression
            with open(file, 'rb') as zf:
                if zf.read(4) == b'\x50\x4B\x03\x04':
                    is_zip = True
    
        if is_gz:
            with gzip.open(file, 'rt') as zf:
                return zf.read()
        elif is_zip:
            #Not supported yet
            return None
    return None

def get_line(file, line_number):
    """
    Return line content for the given line number

    Parameters
    ----------
    file : str
        The input file location or file content.
    line : int
        The line number.

    Returns
    -------
    line
        The line content for the given line number.

    """
    line_count = 0
    is_compressed = check_for_compression(file)
    if os.path.exists(file):

        line_count = 1
        
        if is_compressed:
            data = get_uncompressed_data(file)
            for line in data:
                if line_count == line_number:
                    return line[:-1]
                    break
                line_count += 1
        else:
            with open(file) as f:
                line = f.readline()
                while line:
                    if line_count == line_number:
                        return line[:-1]
                        break
                    line = f.readline()
                    line_count += 1

    return None


def get_col_error_pos(col):
    """
    Return string which points the error in a line

    Parameters
    ----------
    col : int
        Column number where the error is located.

    Returns
    -------
    line : str
        Line with arrow up point to the error location.

    """
    line = "\t"
    i = 1
    while i < col:
        line += " "
        i += 1
    line += "↑"
    return line



# -*- coding: utf-8 -*-

class SigTimingInfo:

    def __init__(self, signal):
        self.signal = signal
        # key is wfc, value is list of tuples (wfe + time offset)
        self.wfc2wfe_time = {}

    def replace_timing_for_wfc(self, wfc, wfe, old_time, new_time):
                
        if wfc in self.wfc2wfe_time:
            pairs = self.wfc2wfe_time[wfc]
            old_value = (wfe, old_time)
            i = pairs.index(old_value) 
            pairs[i] = (wfe, new_time)
        
    def set_timing_for_wfc(self, wfc, wfe, time):

        if wfc in self.wfc2wfe_time:
            pairs = self.wfc2wfe_time[wfc]
            value = (wfe, time)
            pairs.append(value)
        else:
            self.wfc2wfe_time[wfc] = [(wfe, time)]

    def get_wfcs(self):
        '''
        Returns all available WFC for current signal

        Returns
        -------
        List of WFC.

        '''
        return self.wfc2wfe_time.keys()
    
    def get_timing_for_wfc(self, wfc):
        '''
        Returns list of WFE and Timing pair for the given WFC

        Returns
        -------
        List of WFE+Timing.

        '''
        if wfc in self.wfc2wfe_time:
            return self.wfc2wfe_time[wfc]
        else:
            return None
        
    def __str__(self):
        data = f"signal {self.signal} "
        for wfc in self.wfc2wfe_time:
            data += "\nWFC " + wfc + " => " 
            pairs = self.wfc2wfe_time[wfc]
            for p in pairs:
                data += str(p[1]) + ":" + str(p[0]) + " | " 
        return data


# -*- coding: utf-8 -*-
import inspect

from .DomainUtils import DomainUtils
from .WFCUtils import WFCUtils
from .HashInfo import HashInfo
from .STILSemanticException import STILSemanticException


class MacroDefsBlockParser:
    def __init__(self, debug=False):
        self.debug = debug

        self.is_condition_stmt = False
        self.is_first_vector = True

        # List with all macrodefs block names
        self.macrodefs_domains = []

        # dict key   is the macrodefs domain name
        # dict value is a list with all macros defined for this domain
        self.macro_domain2macro_names = {}
        # list with macro names (with domain prefix) which have shift statement
        self.macro_with_shift = []
        # list with macro names (with domain prefix) which have # or % for substitution
        self.macro_for_subs = []

        # dict key   is domain::macro_name
        # dict value is vector addresses
        self.macro2va = {}

        # dict key   is domain::macro_name
        # dict value is test cyclces : ordinary vectors + loops
        self.macro2tc = {}

        # Current domain name
        self.curr_macro_domain = DomainUtils.global_domain
        # current macro name is in full domain format -> domain::macro_name
        self.curr_macro_name = ""
        # Current signal/signal group
        self.curr_sig_ref = ""
        # Current WFC data
        self.curr_vec_data = ""
        # Current WFT
        self.curr_wft = ""
        # Position relative to the shift block
        self.pos_rel_shift = HashInfo.POS_UNKNOWN
        # dict key   is the domain::macro_name
        # dict value is a list with all HashInfo objects
        self.hashinfo = {}

        # dict key   is the macro name in full domain format -> domain::macro_name
        # dict value is a list with signal names found in the first vector address
        self.macro2sig_order = {}

        # List of signals/signal groups for the current vector address
        self.curr_macro_sig_order = []
        # List of WFC for the current vector address
        self.curr_macro_wfc_order = []
        # List of WFC for the current condition statements
        self.curr_macro_cond_wfc_order = []

        # Contains WFC data for signals found in the first Condition statament,
        # before the first vector statement
        # dict key   is the macro name in full domain format -> domain::macro_name
        # dict value is a dict with key signal name and value WFC 
        self.macro2first_cond = {}

        # Start VA of the loop block
        self.start_loop_va = 0
        # Loop count value
        self.loop_count = 0

        self.is_found_hash_macro = False
        
        # dict key   is the macro name in full domain format -> domain::macro_name
        # dict value is a list with all found labels in this procedure
        self.macro_labels = {}
        # dict key   is the macro name in full domain format -> domain::macro_name
        # dict value is a list with all found goto labels in this procedure
        self.macro_goto_labels = {}
        # dict key   is the macro name in full domain format -> domain::macro_name
        # dict value is the goto label object
        self.macro_goto_label_obj = {}

        # List with signals wich have % as WFC. Signal name in full domain format -> domain::signal_name
        self.macro_sig_wfc_percent = []

        # List with signals wich have # as WFC. Signal name in full domain format -> domain::signal_name
        self.macro_sig_wfc_hash = []

    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')

    def b_macrodefs__MACRODEFS_DOMAIN_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_macro_domain = t.value

    def b_macrodefs__OPEN_MACRO_DEFS_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if self.is_signal_block_defined == False:
            err_msg = "Expected Signal block before any MacroDefs block"
            raise Exception(err_msg)
        if self.is_signalgroups_block_defined == False:
            err_msg = "Expected SignalGroups block before any MacroDefs block"
            raise Exception(err_msg)
        if self.is_timing_block_defined == False:
            err_msg = "Expected Timing block before any MacroDefs block"
            raise Exception(err_msg)
        if self.is_patternburst_block_defined == False:
            err_msg = "Expected PatternBurst block before any MacroDefs block"
            raise Exception(err_msg)
        if self.is_patternexec_block_defined == False:
            err_msg = "Expected PatternExec block before any MacroDefs block"
            raise Exception(err_msg)

        if self.curr_macro_domain in self.macrodefs_domains:
            domain_name = DomainUtils.get_domain(self.curr_macro_domain, True)
            err_msg = f"Macrodefs block domain '{domain_name}' is already defined !"
            raise Exception(err_msg)
        else:
            self.macrodefs_domains.append(self.curr_macro_domain)

    def b_macrodefs__MACRO_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        macro_name = t.value
        macro_names = self.macro_domain2macro_names.get(self.curr_macro_domain)
        if macro_names == None:
            macro_names = [macro_name]
            self.macro_domain2macro_names[self.curr_macro_domain] = macro_names
        elif macro_name in macro_names:
            err_msg = f"Macro '{macro_name}' is already defined !"
            raise Exception(err_msg)
        else:
            macro_names.append(macro_name)
        self.curr_macro_name = DomainUtils.get_full_name(self.curr_macro_domain, macro_name)
        self.macro2va[self.curr_macro_name] = 0
        self.macro2tc[self.curr_macro_name] = 0

        self.macro_labels[self.curr_macro_name] = []
        self.macro_goto_labels[self.curr_macro_name] = []
        self.macro_goto_label_obj = {}
        self.macro_sig_wfc_hash = []
        self.macro_sig_wfc_percent = []

    def b_macrodefs__open_macro(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.pos_rel_shift = HashInfo.POS_BEFORE_SHIFT
        self.va = 0

    def b_macrodefs__pattern_statements__USER_KEYWORD_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        
        if t.value not in self.user_defined_keywords:
            err_msg = f"User keyword {t.value} is not defined!"
            raise Exception(err_msg)

    def b_macrodefs__pattern_statements__WAVEFORM_TABLE_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        wft_name = t.value
        self.curr_wft = wft_name

        for pattern_exec_block in self.patt_exec_block2patt_burst.keys():
            patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
            # For every pattern burst block in the pattern exec block
            for patt_burst in patt_bursts:

                macro_domain = self.patt_burst2macro_domain[patt_burst]
                if macro_domain != self.curr_macro_domain:
                    continue

                time_domain = self.patt_exec_block2time_domain[pattern_exec_block]
                wft_list = self.time_domain2wft[time_domain]
                if wft_name not in wft_list:
                    td = DomainUtils.get_domain(time_domain, True)
                    err_msg = f"Wafeform Table named '{wft_name}' from '{td}' timing domain is not defined !"
                    raise Exception(err_msg)

    def b_macrodefs__pattern_statements__KEYWORD_V(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.macro2va[self.curr_macro_name] += 1
        self.macro2tc[self.curr_macro_name] += 1
        self.is_vector_stmt = True

    def b_macrodefs__pattern_statements__KEYWORD_VECTOR(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.macro2va[self.curr_macro_name] += 1
        self.macro2tc[self.curr_macro_name] += 1
        self.is_vector_stmt = True

    def b_macrodefs__pattern_statements__LABEL(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if t.value[-1] == ':':
            label_str = t.value[0:-1]
        else:
            label_str = t.value
        label_strip = label_str.strip()
        if label_strip.startswith('"') == False and label_strip.endswith('"') == False:
            label_split = label_strip.split(" ")
            label = label_split[0].strip()
        else:
            label = label_strip

        labels = self.macro_labels[self.curr_macro_name]
        if label in labels:
            err_msg = f"Label '{label}' is already defined !"
            raise Exception(err_msg)
        else:
            labels.append(label)

        return label
            

    def b_macrodefs__pattern_statements__GOTO_LABEL(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        gt_label = t.value
        labels = self.macro_goto_labels[self.curr_macro_name]
        if gt_label not in labels:
            labels.append(gt_label)
        
        self.macro_goto_label_obj[gt_label] = t

    def b_macrodefs__pattern_statements__SIGREF_EXPR(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_sig_ref = t.value

        if self.curr_sig_ref in self.sig2type:
            # signal ref is defined in the Signal block
            pass
        else:
            # check if sigref is already defined signal group
            # get all signal groups
            if self.curr_sig_ref not in self.all_signal_groups:
                err_msg = f"Signal '{self.curr_sig_ref}' is not defined !"
                raise Exception(err_msg)

        self.curr_macro_sig_order.append(self.curr_sig_ref)

    def b_macrodefs__pattern_statements__VEC_DATA_STRING(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        vec_data = WFCUtils.expand_wfcs(t.value)
        
        if self.is_vector_stmt == False:
            self.curr_macro_cond_wfc_order.append(vec_data)

        # Collect increamental substitution information
        pos = WFCUtils.find_hash(vec_data)
        if len(pos) > 0:
            hi = HashInfo(self.pos_rel_shift, self.curr_sig_ref, pos)
            hid = self.hashinfo.get(self.curr_macro_name)
            if hid == None:
                self.hashinfo[self.curr_macro_name] = [hi]
            else:
                hid.append(hi)

        # Check number of WFC against signal/signalgroups
        if self.curr_sig_ref in self.sig2type:
            if len(vec_data) > 1:
                err_msg = f"Signal '{self.curr_sig_ref}' has more than one WFC !"
                raise Exception(err_msg)
            sig = self.curr_sig_ref
            if self.curr_macro_domain in self.macro_domain2timing_domain:
                tds = self.macro_domain2timing_domain[self.curr_macro_domain]
                for d in tds:
                    indx = 0
                    fstn = DomainUtils.get_full_name(d, sig)
                    if fstn not in self.sig2wfc:
                        continue
                    wfc_list = self.sig2wfc[fstn]
                    wfc = vec_data[indx : indx + 1]
                    if wfc == "%":
                        if sig in self.macro_sig_wfc_hash:
                            err_msg = f"Signal '{sig}' can not have both '%' and '#' as WFC!"
                            raise Exception(err_msg)
                        elif sig not in self.macro_sig_wfc_percent:
                            self.macro_sig_wfc_percent.append(sig)
                        fmn = self.curr_macro_name
                        if fmn not in self.macro_for_subs:
                            self.macro_for_subs.append(fmn)
                    elif wfc == "#":
                        if sig in self.macro_sig_wfc_percent:
                            err_msg = f"Signal '{sig}' can not have both '%' and '#' as WFC!"
                            raise Exception(err_msg)
                        elif sig not in self.macro_sig_wfc_hash:
                            self.macro_sig_wfc_hash.append(sig)
                        fmn = self.curr_macro_name
                        if fmn not in self.macro_for_subs:
                            self.macro_for_subs.append(fmn)
                        self.is_found_hash_macro = True
                    elif wfc not in wfc_list:
                        dn = DomainUtils.get_domain(d, True)
                        err_msg = f"Not defined WFC '{wfc}' for signal {sig} for the '{dn}' domain!"
                        raise Exception(err_msg)
                    indx += 1
        else:
            for pattern_exec_block in self.patt_exec_block2patt_burst.keys():
                patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
                # For every pattern burst block in the pattern exec block
                for patt_burst in patt_bursts:
                    sig_group_domain = self.patt_burst2sig_groups_domain[patt_burst]

                    macro_domain = self.patt_burst2macro_domain[patt_burst]

                    if macro_domain != self.curr_macro_domain:
                        continue

                    fsg = DomainUtils.get_full_name(sig_group_domain, self.curr_sig_ref)
                    domain = DomainUtils.get_domain(fsg, True)

                    if fsg in self.signal_groups2signals:
                        signals = self.signal_groups2signals[fsg]
                        if len(signals) != len(vec_data):
                            #pass
                            err_msg = f"Number of WFC ({len(vec_data)}) does not match with the number of signals ({len(signals)}) for signal groups '{self.curr_sig_ref}' in {domain} domain!"
                            raise Exception(err_msg)
                        else:
                            # Check if WFC are defined for the signal
                            td = self.patt_exec_block2time_domain[pattern_exec_block]
                            indx = 0
                            for sig in signals:
                                fstn = DomainUtils.get_full_name(td, sig)
                                wfc_list = self.sig2wfc[fstn]
                                wfc = vec_data[indx : indx + 1]
                                if wfc == "%":
                                    if sig in self.macro_sig_wfc_hash:
                                        err_msg = f"Signal '{sig}' can not have both '%' and '#' as WFC!"
                                        raise Exception(err_msg)
                                    elif sig not in self.macro_sig_wfc_percent:
                                        self.macro_sig_wfc_percent.append(sig)
                                    fmn = self.curr_macro_name
                                    if fmn not in self.macro_for_subs:
                                        self.macro_for_subs.append(fmn)
                                elif wfc == "#":
                                    if sig in self.macro_sig_wfc_percent:
                                        err_msg = f"Signal '{sig}' can not have both '%' and '#' as WFC!"
                                        raise Exception(err_msg)
                                    elif sig not in self.macro_sig_wfc_hash:
                                        self.macro_sig_wfc_hash.append(sig)
                                    fmn = self.curr_macro_name
                                    if fmn not in self.macro_for_subs:
                                        self.macro_for_subs.append(fmn)
                                    self.is_found_hash_macro = True
                                elif wfc not in wfc_list:
                                    sig_group_name = DomainUtils.get_name(fsg)
                                    timing_domain = DomainUtils.get_domain(td, True)
                                    err_msg = f"Not defined WFC '{wfc}' for signal {sig} in the signal groups '{sig_group_name}' for the '{timing_domain}' timing domain!"
                                    raise Exception(err_msg)
                                indx += 1
                    else:
                        err_msg = f"Signal or SignalGroup {self.curr_sig_ref} is not defined in Signals group or in SignalGroups {domain} domain!"
                        raise Exception(err_msg)
        self.curr_macro_wfc_order.append(vec_data)

    def b_macrodefs__pattern_statements__CLOSE_VECTOR_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_sig_ref = ""

        # Getting condition WFC data before the first vector statement
        if self.is_condition_stmt and self.is_first_vector:
            for pattern_exec_block in self.patt_exec_block2patt_burst.keys():
                patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
                # For every pattern burst block in the pattern exec block
                for patt_burst in patt_bursts:
                    sig_group_domain = self.patt_burst2sig_groups_domain[patt_burst]
                    sig2wfc = WFCUtils.collect_sig2wfc(self.sig2type.keys(),
                                                       sig_group_domain,
                                                       self.signal_groups2signals,
                                                       self.curr_macro_sig_order,
                                                       self.curr_macro_wfc_order)
                    self.macro2first_cond[self.curr_macro_name] = sig2wfc

 

        #Ignore so far the Fixed and Condition statetments        
        if self.is_vector_stmt:
#            self.save_cmd_macro()
            if self.is_first_vector:
                self.is_first_vector = False
            self.curr_wft = ""

        self.curr_macro_sig_order = []
        self.curr_macro_wfc_order = []
        self.curr_macro_cond_wfc_order = []
        self.is_vector_stmt = False

    def b_macrodefs__pattern_statements__KEYWORD_SHIFT(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.macro_with_shift.append(self.curr_macro_name)
        self.pos_rel_shift = HashInfo.POS_SHIFT

    def b_macrodefs__pattern_statements__close_shift_block(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.pos_rel_shift = HashInfo.POS_AFTER_SHIFT
        self.is_found_hash_macro = False

    def b_macrodefs__pattern_statements__CALL_VEC_DATA_STRING(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

    def b_macrodefs__close_macro(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        macro_name = DomainUtils.get_name(self.curr_macro_name)
        labels = self.macro_labels[self.curr_macro_name]
        gt_labels = self.macro_goto_labels[self.curr_macro_name]
        for label in gt_labels:
            if label not in labels:
                obj = self.macro_goto_label_obj[label]
                err_msg = f"Label {label} used by Goto statament in the Macro '{macro_name}' is not defined!"
                raise STILSemanticException( obj.line, obj.column, err_msg)

        self.curr_macro_name = ""
        self.pos_rel_shift = HashInfo.POS_UNKNOWN
        self.is_first_vector = True

    def b_macrodefs__pattern_statements__LOOP_COUNT(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        if int(t.value) < 1:
            err_msg = "Loop value must be positive!"
            raise Exception(err_msg)
        
        self.start_loop_va = self.macro2va[self.curr_macro_name]
        self.loop_count = int(t.value)

    def b_macrodefs__pattern_statements__close_loop_block(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        # Used to compare test cycles with ATPG patterns files.
        curr_va = self.macro2va[self.curr_macro_name] - 1
        v = (curr_va - self.start_loop_va + 1) * (self.loop_count - 1)
        self.macro2tc[self.curr_macro_name] += v
        self.start_loop_va = 0
        self.loop_count = 0

    def b_macrodefs__pattern_statements__MATCHLOOP_COUNT(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        if int(t.value) < 1:
            err_msg = "MatchLoop value must be positive!"
            raise Exception(err_msg)

    def b_macrodefs__pattern_statements__close_matchloop_block(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

    def b_macrodefs__CLOSE_MACRO_DEFS_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
            for key in self.macro_domain2macro_names.keys():
                macro_names = self.macro_domain2macro_names.get(key)
                for name in macro_names:
                    print(f" macro domain {key} -> macro name {name}")
        self.curr_macro_domain = DomainUtils.global_domain



# -*- coding: utf-8 -*-
import re

from .DomainUtils import DomainUtils


class WFCUtils:

    debug = False

    def __init__(self):
        pass

    def trim_wfcs(wfc_list):
        if WFCUtils.debug:
            print("WFCUtils::trim_wfcs")
        return re.sub(r"\s+", "", wfc_list)

    def expand_wfcs(wfc_list):
        """
        This function will expand WFC sequence in single continious string
        Useful in the case when macro/procedure are called with signal's WFC

        LH 0 1         -> LH01
        LH \r2 01 XX X -> LH0101XXX
        """
        if WFCUtils.debug:
            print("WFCUtils::expand_wfcs")

        wfcs = ""
        
        # Check if the WFCs list contain "\j" join command
        join_pos = wfc_list.find("\\j")
        if join_pos != -1:
            wfc_list = wfc_list[join_pos+2:]

        # Check if the WFCs list contain "\m" map command
        map_pos = wfc_list.find("\\m")
        if map_pos != -1:
            wfc_list = wfc_list[map_pos+2:]

        # Check if the WFCs list contain "\r" repeat command
        repeat_pos = wfc_list.find("\\r")
        if repeat_pos == -1:
            # no repeat commans, just remove all white spaces
            wfcs = re.sub(r"\s+", "", wfc_list)
        else:
            start_pos = 0
            # collect all WFC before the first /r command
            wfcs += wfc_list[start_pos:repeat_pos]
            while start_pos != -1:

                if repeat_pos == -1:
                    # the last segment
                    repeat_list = wfc_list[start_pos:]
                else:
                    # get one segment with /r command
                    repeat_list = wfc_list[start_pos:repeat_pos]
                # split by empty space to colect the repeat value /rXXX
                parts = repeat_list.split(" ")

                repeat_value = parts[0][2:]
                if len(repeat_value) > 0:
                    r = int(repeat_value.strip(), base=10)
                    for i in range(r):
                        # repeat the WFCs "r" times
                        wfcs += parts[1]
                elif parts[0].startswith("\\r"):
                    raise Exception("Missing repeat value after \\r command")
                # Adding rest of the WFCs after repeat statement
                if len(parts) > 2:
                    for i in range(len(parts) - 2):
                        wfcs += parts[i + 2]
                if repeat_pos == -1:
                    start_pos = -1
                else:
                    # search for the next repeat command
                    start_pos = repeat_pos
                    repeat_pos = wfc_list.find("\\r", repeat_pos + 1)
            wfcs = re.sub(r"\s+", "", wfcs)

        return wfcs

    def collect_sig2wfc(signals, 
                        sig_group_domain,
                        signals_group2signals,
                        curr_signals,
                        curr_wfcs,
                        last_sig2wfc = None,
                        sig2wfc_before_subs = None
                        ):
        """
        This function will collect WFC for every signal in a vector or condition
        statement. If last_sig2wfc is given, it will fill missing sig2wfc
        Will return a dict with signal name as key and WFC as value.
        signals -> list with signal names
        sig_group_domain -> signals group domain name
        signals_group2signals -> dict with key domain::signals_group and key list of signals names
        curr_signals -> list with signals or signals groups
        curr_wfcs -> list with single (per signal) or multiple (per signals group) WFC
        last_sig2wfc -> last known state of signals and their WFC (optional). 
                        Key is the signal name, value is a WFC
        sig2wfc_before_subs -> last known WFC for signals before first # or %
        """        
        if WFCUtils.debug:
            print("WFCUtils::collect_sig2wfc")
            print(f"WFCUtils::collect_sig2wfc signals {signals}")
            print(f"WFCUtils::collect_sig2wfc sig_group_domain {sig_group_domain}")
            print(f"WFCUtils::collect_sig2wfc signals_group2signals {signals_group2signals}")
            print(f"WFCUtils::collect_sig2wfc curr_signals {curr_signals}")
            print(f"WFCUtils::collect_sig2wfc curr_wfcs {curr_wfcs}")
            print(f"WFCUtils::collect_sig2wfc last_sig2wfc {last_sig2wfc}")
            print(f"WFCUtils::collect_sig2wfc sig2wfc_before_subs {sig2wfc_before_subs}")

        sig2wfc = {}
        found_sigs = []
        sig_indx = 0
        for sig in curr_signals:

            sg = DomainUtils.get_full_name(sig_group_domain, sig)

            if sig in signals:
                wfc = curr_wfcs[sig_indx]
                sig2wfc[sig] = [wfc]
                sig_indx += 1
                found_sigs.append(sig)
                
                if wfc == '#' or wfc == '%':
                    if sig not in sig2wfc_before_subs:
                        last_wfc = last_sig2wfc[sig][0]
                        sig2wfc_before_subs[sig] = last_wfc

                if WFCUtils.debug:
                    print(f"WFCUtils::collect_sig2wfc signal {sig} wfc {wfc}")

            elif sg in signals_group2signals:
                si = 0
                for sigg in signals_group2signals[sg]:
                    wfc = curr_wfcs[sig_indx]
                    #print(f" wfc {wfc} for sig index {sig_indx} signal {sigg}")
                    sig2wfc[sigg] = [wfc[si]]
                    found_sigs.append(sigg)
                    si += 1

                    """
                    In the standard is not specified exactly which WFC should 
                    be used in delta vector after WFC substitution like:
                        
                    V {sig = WFC}// Last defined WFC before WFC substitution
                    V {sig = #}  // WFC substitution
                    V {}         // Delta vector, which WFC should be placed here?
                    
                    The implementation will set in the delta vector, the last 
                    defined WFC before the first substituted WFC (# or %). 
                    This is based 
                    
                    """
                    if wfc == '#' or wfc == '%':
                        last_wfc = last_sig2wfc[sigg][0]
                        sig2wfc_before_subs[sigg] = last_wfc

                    if WFCUtils.debug:
                        print(f"WFCUtils::collect_sig2wfc signal {sigg} wfc {sig2wfc[sigg]}")
                sig_indx += 1
            else:
                raise Exception(f"Error can not find signal/signal group named {sig}")
        
        # Fill with missing wfc and signals
        if last_sig2wfc != None:
            for sig in last_sig2wfc:
                if sig not in found_sigs:
                    last_wfc = last_sig2wfc[sig][0]
                    if last_wfc == "#" or last_wfc == "%":
                        if sig2wfc_before_subs == None:
                            err = f"For signal {sig}, there is no defined WFC before first '#' or '%'"
                            raise Exception(err)
                        else:
                            if sig in sig2wfc_before_subs:
                                sig2wfc[sig] = [sig2wfc_before_subs[sig]]
                            else:
                                # Issue #13
                                # At this time is unknown which WFC should be placed
                                # When Macro/Proc call arguments are known, the '?' 
                                # should be replaced with the last used WFC in '#'
                                sig2wfc[sig] = '?'
                    else:
                        sig2wfc[sig] = [last_sig2wfc[sig][0]]
                
        return sig2wfc

    def collect_va_wfcs(
        signal_groups_domains,
        pattern,
        patt2sig_order,
        curr_sig_order,
        curr_wfc_order,
        last_wfc_order,
        signals,
        signal_groups2signals,
    ):
        """
        This function will collect WFC sequence for a vector address,
        based on signal order and WFC sequence from the last vector.

        V {sig1 = 0; sig2 = L; sig3 = X;} // VA=0 -> 0LX
        V {sig3 = H;}                     // VA=1 -> 0LH
        The VA=1 does not contain all required WFC, so according
        STIL Standard, the latest defined WFC (at VA=0) will be used.

        curr_sig_order -> list with signal/SignalsGroup for the current VA
        curr_wfc_order -> list with WFC for the current VA
        last_wfc_order -> WFCs in the last VA
        signal_order   -> list with signal order in the first VA of the pattern block

        """
        if WFCUtils.debug == False:
            print("WFCUtils::collect_va_wfcs")
            print(f"WFCUtils::collect_va_wfcs curr_sig_order {curr_sig_order}")
            print(f"WFCUtils::collect_va_wfcs curr_wfc_order {curr_wfc_order}")
            print(f"WFCUtils::collect_va_wfcs last_wfc_order {last_wfc_order}")

        wfcs = ""

        for sp in signal_groups_domains:
            key = DomainUtils.get_full_name(sp, pattern)
            print(f"key {key}")
            if key in patt2sig_order:
                sig_order = patt2sig_order[key]
                print(f"sig_order {sig_order} for {key}")

                si = 0
                for sig_o in sig_order:
                    print(f"Process signal {sig_o}")
                    for sigs in curr_sig_order:
                        print(f"\t compare to {sigs}")
                        if sigs == sig_o:
                            i = curr_sig_order.index(sigs)
                            wfc = curr_wfc_order[i]
                            print(f"s signal {sigs} wfc {wfc}")
                            wfcs += wfc
                            break
                        else:
                            for sgd in signal_groups_domains:
                                key = DomainUtils.get_full_name(sgd, sigs)
                                if key in signal_groups2signals:
                                    signals = signal_groups2signals[key]
                                    for sig in signals:
                                        if sig == sig_o:
                                            wfc = curr_wfc_order[0][si]
                                            wfcs += wfc
                                            print(f"sg signal {sig} wfc {wfc}")
                                            si += 1
                                            break

        if WFCUtils.debug == False:
            print(f"WFCUtils::collect_va_wfcs return value {wfcs}")

        return wfcs

    def find_hash(wfc_list):
        """
        This function will return position of all # in a list of WFCs

        """
        plist = []
        pos = wfc_list.find("#")
        while pos != -1:
            plist.append(pos)
            pos = wfc_list.find("#", pos + 1)
        return plist



# -*- coding: utf-8 -*-
import inspect

from lark.exceptions import VisitError
from .WFCUtils import WFCUtils
from .DomainUtils import DomainUtils
from .STILSemanticException import STILSemanticException

class PatternBlockParser:
    def __init__(self, debug=False):
        self.debug = debug

        self.is_first_pattern = False
        self.is_first_vector = True

        # List with all pattern names
        self.patterns = []
        # Name of the current pattern block
        self.curr_pattern = ""
        # dict key   is the pattern block name
        # dict value is a list with procedure names (without domain name) which are called in this pattern block
        self.patt2call = {}
        # dict key   is the pattern block name
        # dict value is a list with macro names (without domain name) which are called in this pattern block
        self.patt2macro = {}

        # Current signal/signal group
        self.curr_sig_ref = ""
        # Current WFC data
        self.curr_vec_data = ""
        # Current WFT
        self.curr_wft = ""

        self.curr_macro_name_call = ""
        self.curr_proc_name_call = ""

        # dict key   is the pattern name together with signals group domain -> SignalsGroupDomain::pattern_name
        # dict value is a list with signal names found in the first vector address
        self.patt2sig_order = {}

        # List of signals/signal groups for the current vector address
        self.curr_sig_order = []
        # List of WFC for the current vector address
        self.curr_wfc_order = []

        # dict key   is the pattern name
        # dict value is a list with all found labels in this pattern
        self.patt_labels = {}
        # dict key   is the pattern name
        # dict value is a list with all found goto labels in this pattern
        self.patt_goto_labels = {}
        # dict key   is the label name of the goto statement
        # dict value is the goto label object
        self.goto_label_obj = {}

    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')


    def b_pattern__pattern_statements__USER_KEYWORD_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        
        if t.value not in self.user_defined_keywords:
            err_msg = f"User keyword {t.value} is not defined!"
            raise Exception(err_msg)
        
    def b_pattern__PATTERN_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if t in self.patterns:
            err_msg = f"Pattern block {t} already defined! Pattern block names must be unique."
            raise Exception(err_msg)
        else:
            self.curr_pattern = t.value
            self.patterns.append(self.curr_pattern)
            self.patt2call[self.curr_pattern] = []
            self.patt2macro[self.curr_pattern] = []
            self.patt_labels[self.curr_pattern] = []
            self.patt_goto_labels[self.curr_pattern] = []
            self.goto_label_obj = {}

    def b_pattern__OPEN_PATTERN_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if self.is_signal_block_defined == False:
            err_msg = "Expected Signal block before any Pattern block"
            raise Exception(err_msg)
        if self.is_signalgroups_block_defined == False:
            err_msg = "Expected SignalGroups block before any Pattern block"
            raise Exception(err_msg)
        if self.is_timing_block_defined == False:
            err_msg = "Expected Timing block before any Pattern block"
            raise Exception(err_msg)
        if self.is_patternburst_block_defined == False:
            err_msg = "Expected PatternBurst block before any Pattern block"
            raise Exception(err_msg)
        if self.is_patternexec_block_defined == False:
            err_msg = "Expected PatternExec block before any Pattern block"
            raise Exception(err_msg)

        if self.is_first_pattern == False:
            self.is_first_pattern = True

            """
            # Check if all macro/procedures domains used in PatternBurst are defined
            # For every pattern exec block
            for pattern_exec_block in self.patt_exec_block2patt_burst.keys():
                patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
                # For every pattern burst block in the pattern exec block
                for patt_burst in patt_bursts:
                    proc_domain = self.patt_burst2proc_domain[patt_burst]
                    if proc_domain not in self.proc_domain2proc_names:
                        err_msg = f"Procedure domain '{proc_domain}' is not defined in pattern burst {patt_burst}"
                        raise Exception(err_msg)
                    macro_domain = self.patt_burst2macro_domain[patt_burst]
                    if macro_domain not in self.macro_domain2macro_names:
                        print(
                            f"macro_domain {macro_domain} {self.macro_domain2obj} {self.macro_domain2obj[macro_domain].line}"
                        )
                        err_msg = f"MacroDefs domain '{macro_domain}' used in the pattern burst block {patt_burst} is not defined !"
                        raise VisitError(
                            err_msg,
                            self.macro_domain2obj[macro_domain],
                            Exception(err_msg),
                        )
            """
        self.va = 0

    def b_pattern__pattern_statements__WAVEFORM_TABLE_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        wft_name = t.value
        self.curr_wft = wft_name

        for pattern_exec_block in self.patt_exec_block2patt_burst.keys():
            patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
            # For every pattern burst block in the pattern exec block
            for patt_burst in patt_bursts:

                pattern_list = self.patt_burst_block2pattern_blocks[patt_burst]

                # For every pattern block in the pattern burst block
                for pattern in pattern_list:

                    if pattern != self.curr_pattern:
                        continue

                    time_domain = self.patt_exec_block2time_domain[pattern_exec_block]
                    wft_list = self.time_domain2wft[time_domain]
                    if wft_name not in wft_list:
                        td = DomainUtils.get_domain(time_domain, True)
                        err_msg = f"Wafeform Table named '{wft_name}' from '{td}' timing domain is not defined !"
                        raise Exception(err_msg)

    def b_pattern__pattern_statements__KEYWORD_V(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.is_vector_stmt = True

    def b_pattern__pattern_statements__KEYWORD_VECTOR(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.is_vector_stmt = True

    def b_pattern__pattern_statements__LABEL(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if t.value[-1] == ':':
            label_str = t.value[0:-1]
        else:
            label_str = t.value
        label_strip = label_str.strip()
        if label_strip.startswith('"') == False and label_strip.endswith('"') == False:
            label_split = label_strip.split(" ")
            label = label_split[0].strip()
        else:
            label = label_strip
            
        labels = self.patt_labels[self.curr_pattern]
        if label in labels:
            err_msg = f"Label '{label}' is already defined !"
            raise Exception(err_msg)
        else:
            labels.append(label)

        return label

    def b_pattern__pattern_statements__GOTO_LABEL(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        gt_label = t.value
        labels = self.patt_goto_labels[self.curr_pattern]
        if gt_label not in labels:
            labels.append(gt_label)
        
        self.goto_label_obj[gt_label] = t

    def b_pattern__pattern_statements__SIGREF_EXPR(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_sig_ref = t.value

        self.curr_sig_order.append(self.curr_sig_ref)

    def b_pattern__pattern_statements__VEC_DATA_STRING(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        vec_data = WFCUtils.expand_wfcs(t.value)

        self.curr_wfc_order.append(vec_data)

        sigref = self.curr_sig_ref

        if self.curr_sig_ref in self.sig2type:
            # Check if WFC is defined for the signal
            if len(vec_data) > 1:
                err_msg = f"Signal '{self.curr_sig_ref}' has more than one WFC !"
                raise Exception(err_msg)
            sig = self.curr_sig_ref
            if self.curr_pattern in self.patt2timing_domain:
                tds = self.patt2timing_domain[self.curr_pattern]
                for d in tds:
                    indx = 0
                    dn = DomainUtils.get_domain(d, True)
                    fstn = DomainUtils.get_full_name(d, sig)
                    if fstn in self.sig2wfc:
                        wfc_list = self.sig2wfc[fstn]
                        wfc = vec_data[indx : indx + 1]
                        if wfc not in wfc_list:
                            err_msg = f"Not defined WFC '{wfc}' for signal {sig} for the '{dn}' domain!"
                            raise Exception(err_msg)
                        indx += 1
                    else:
                        pass
                        # ToDo!!!
                        # err_msg = f"Signal '{sig}' does not have WFCs defined in timing domain '{dn}'!"
                        # raise Exception(err_msg)
        else:
            # check if sigref is already defined signal group

            if self.curr_pattern in self.used_patterns:

                if self.curr_pattern in self.patt2sig_group_domain:

                    sig_group_domains = self.patt2sig_group_domain[self.curr_pattern]

                    is_signal_group_found = False

                    for sig_group_domain in sig_group_domains:

                        sig_group_name = DomainUtils.get_full_name(
                            sig_group_domain, sigref
                        )
                        sig_groups = self.signal_groups2signals.keys()

                        for sig_group in sig_groups:
                            sgd = DomainUtils.get_domain(sig_group)
                            if sgd != sig_group_domain:
                                continue
                            sig_group_name = DomainUtils.get_name(sig_group)
                            if sig_group_name == sigref:
                                is_signal_group_found = True
                                signals = self.signal_groups2signals.get(sig_group)
                                # check if signal length corresponds to the WFC length
                                if len(signals) != len(vec_data):
                                    domain = DomainUtils.get_domain(sig_group, True)
                                    err_msg = f"Number of WFC ({len(vec_data)}) does not match with the number of signals ({len(signals)}) for signal groups '{sig_group_name}' in {domain} domain!"
                                    raise Exception(err_msg)
                                else:
                                    # Check if WFC are defined for the signal
                                    if self.curr_pattern in self.patt2timing_domain:
                                        tds = self.patt2timing_domain[self.curr_pattern]
                                        for d in tds:
                                            indx = 0
                                            for sig in signals:
                                                fstn = DomainUtils.get_full_name(d, sig)
                                                if fstn not in self.sig2wfc:
                                                    dn = DomainUtils.get_domain(d, True)
                                                    err_msg = f"Signal {sig} is not defined in the {dn} Timing block!"
                                                    raise Exception(err_msg)
                                                wfc_list = self.sig2wfc[fstn]
                                                wfc = vec_data[indx : indx + 1]
                                                if wfc not in wfc_list:
                                                    dn = DomainUtils.get_domain(d, True)
                                                    err_msg = f"Not defined WFC '{wfc}' for signal {sig} in the signal groups '{sig_group_name}' for the '{dn}' domain!"
                                                    raise Exception(err_msg)
                                                indx += 1
                    if is_signal_group_found == False:
                        err_msg = f"Signal/Signal group {sigref} is not defined!"
                        raise Exception(err_msg)
                            

                else:
                    # pattern is not used
                    pass

    def b_pattern__pattern_statements__CLOSE_VECTOR_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        
        #Ignore so far the Fixed and Condition statetments        
        if self.is_vector_stmt:

            # Collect signal order for the first vector address
            if self.is_first_vector:
                self.is_first_vector = False
                # print(f"\n=> {self.curr_pattern}")
                if self.curr_wft == "":
                    err_msg = (
                        "Missing WaveformTable reference before the first vector address!"
                    )
                    raise Exception(err_msg)
            self.curr_wft = ""

        self.curr_sig_order = []
        self.curr_wfc_order = []

        self.curr_sig_ref = ""
        self.curr_vec_data = ""
        self.is_vector_stmt = False

    def b_pattern__pattern_statements__CALL_PROC_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.patt2call[self.curr_pattern].append(t.value)
        self.curr_proc_name_call = t.value

        # Check if procedure is defined for all procedures domains
        for pattern_exec_block in self.patt_exec_block2patt_burst.keys():
            patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
            # For every pattern burst block in the pattern exec block
            for patt_burst in patt_bursts:
                pattern_list = self.patt_burst_block2pattern_blocks[patt_burst]
                # For every pattern block in the pattern burst block
                for pattern in pattern_list:
                    proc_domain = self.patt_burst2proc_domain[patt_burst]
                    if proc_domain in self.proc_domain2proc_names:
                        proc_names_defines = self.proc_domain2proc_names[proc_domain]
                        if pattern in self.patt2call:
                            proc_calls = self.patt2call[pattern]
                            for proc_name_call in proc_calls:

                                if proc_name_call not in proc_names_defines:

                                    if proc_domain == DomainUtils.global_domain:
                                        err_msg = f"Procedure {proc_name_call} called in the pattern {pattern} is not defined in the global procedure domain !"
                                    else:
                                        err_msg = f"Procedure {proc_name_call} called in the pattern {pattern} is not defined in the procedure domain {proc_domain} !"
                                    raise Exception(err_msg)
                    else:
                        if proc_domain == DomainUtils.global_domain:
                            err_msg = f"Unnamed (global) Procedures block is not defined for Procedure named '{self.curr_proc_name_call}' called in the pattern {pattern} !"
                        else:
                            line = self.proc_domain2obj[proc_domain].line
                            err_msg = f"Procedures block '{proc_domain}' in line {line} is not defined for Procedure named '{self.curr_proc_name_call}' called in the pattern {pattern} !"
                        raise Exception(err_msg)

    def b_pattern__pattern_statements__CALL_MACRO_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.patt2macro[self.curr_pattern].append(t.value)
        self.curr_macro_name_call = t.value

        # Check if macro is defined for all macro domains
        for pattern_exec_block in self.patt_exec_block2patt_burst.keys():
            patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
            # For every pattern burst block in the pattern exec block
            for patt_burst in patt_bursts:
                pattern_list = self.patt_burst_block2pattern_blocks[patt_burst]
                # For every pattern block in the pattern burst block
                for pattern in pattern_list:
                    macro_domain = self.patt_burst2macro_domain[patt_burst]
                    if macro_domain in self.macro_domain2macro_names:
                        macro_names_defines = self.macro_domain2macro_names[
                            macro_domain
                        ]
                        if pattern in self.patt2macro:
                            macro_calls = self.patt2macro[pattern]

                            for macro_name_call in macro_calls:
                                if macro_name_call not in macro_names_defines:
                                    if macro_domain == DomainUtils.global_domain:
                                        err_msg = f"Macro {macro_name_call} called in the pattern {pattern} is not defined in the unnamed (global) macrodefs domain !"
                                    else:
                                        err_msg = f"Macro {macro_name_call} called in the pattern {pattern} is not defined in the macrodefs domain {macro_domain} !"
                                    raise Exception(err_msg)

                    else:
                        if macro_domain == DomainUtils.global_domain:
                            err_msg = f"Unnamed (global) Macrodefs block is not defined for Macro named '{self.curr_macro_name_call}' called in the pattern {pattern} !"
                        else:
                            line = self.macro_domain2obj[macro_domain].line
                            err_msg = f"Macrodefs block '{macro_domain}' in line {line} is not defined for Macro named '{self.curr_macro_name_call}' called in the pattern {pattern} !"
                        raise Exception(err_msg)

    def b_pattern__pattern_statements__CALL_SIGREF_EXPR(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_sig_ref = t.value
        # print(f"\n===> CALL_SIGREF_EXPR {t.value}\n")

        sigref = self.curr_sig_ref

        if sigref in self.sig2type:
            # signal ref is defined in the Signal block
            pass
        else:
            # check if sigref is already defined signal group
            # get all signal groups domain for the current pattern
            sig_group_domains = self.patt2sig_group_domain[self.curr_pattern]
            for sig_group_domain in sig_group_domains:

                sig_group_name = DomainUtils.get_full_name(sig_group_domain, sigref)
                sig_groups = self.signal_groups2signals.keys()

                if sig_group_name not in sig_groups:
                    if sig_group_domain == DomainUtils.global_domain:
                        sig_group_domain = "unnamed"
                    err_msg = f"In the '{sig_group_domain}' signal group domain, there is no definition for signal reference '{self.curr_sig_ref}' used in the pattern block '{self.curr_pattern}' !"
                    raise Exception(err_msg)

    def b_pattern__pattern_statements__CALL_VEC_DATA_STRING(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_sig_ref = ""

    def b_pattern__pattern_statements__MACRO_SIGREF_EXPR(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_sig_ref = t.value

    def b_pattern__pattern_statements__MACRO_VEC_DATA_STRING(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_sig_ref = ""

    def b_pattern__pattern_statements__LOOP_COUNT(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        if int(t.value) < 1:
            err_msg = "Loop value must be positive!"
            raise Exception(err_msg)

    def b_pattern__pattern_statements__close_loop_block(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

    def b_pattern__pattern_statements__MATCHLOOP_COUNT(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        if int(t.value) < 1:
            err_msg = "MatchLoop value must be positive!"
            raise Exception(err_msg)
        
    def b_pattern__pattern_statements__MATCHLOOP_INF(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        

    def b_pattern__pattern_statements__close_matchloop_block(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

    def b_pattern__CLOSE_PATTERN_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        
        labels = self.patt_labels[self.curr_pattern]
        gt_labels = self.patt_goto_labels[self.curr_pattern]
        for label in gt_labels:
            if label not in labels:
                obj = self.goto_label_obj[label]
                err_msg = f"Label '{label}' used by Goto statament in the Pattern block '{self.curr_pattern}' is not defined!"
                raise STILSemanticException( obj.line, obj.column, err_msg)
                

        self.curr_pattern = ""
        self.is_first_vector = True

    def eof(self):
        if self.debug:
            print("On end of the file")

        # Check if all pattern blocks used in pattern burst are defined
        # For every pattern exec block
        for pattern_exec_block in self.patt_exec_block2patt_burst.keys():
            patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
            # For every pattern burst block in the pattern exec block
            for patt_burst in patt_bursts:
                pattern_list = self.patt_burst_block2pattern_blocks[patt_burst]
                # For every pattern block in the pattern burst block
                for pattern in pattern_list:
                    if pattern not in self.patterns:
                        err_msg = f"Pattern block {pattern} used in pattern burst block {patt_burst} is not defined!"
                        raise VisitError(
                            err_msg, self.patt2obj[pattern], Exception(err_msg)
                        )
                # check domain names for macrodefs, procedures and signal groups
                if patt_burst in self.patt_burst2macro_domain:
                    pass
                if patt_burst in self.patt_burst2proc_domain:
                    pass
                if patt_burst in self.patt_burst2sig_groups_domain:
                    pass



# Copyright (c) Semi-ATE
# Distributed under the terms of the GPLv2 License

from .STILBlockParser import STILBlockParser
from .HeaderBlockParser import HeaderBlockParser
from .SignalsBlockParser import SignalsBlockParser
from .SignalGroupsBlockParser import SignalGroupsBlockParser
from .TimingBlockParser import TimingBlockParser
from .ScanStructBlockParser import ScanStructBlockParser
from .SelectorBlockParser import SelectorBlockParser
from .SpecBlockParser import SpecBlockParser
from .MacroDefsBlockParser import MacroDefsBlockParser
from .ProceduresBlockParser import ProceduresBlockParser
from .PatternBurstBlockParser import PatternBurstBlockParser
from .PatternExecBlockParser import PatternExecBlockParser
from .PatternBlockParser import PatternBlockParser
from .SyntaxParserExceptions import SyntaxParserExceptions
from .STILSemanticException import STILSemanticException
from .STILDumpCompiler import STILDumpCompiler
from .STILDumpCompilerExceptions import STILDumpCompilerException
from .STILDumpCompilerExceptions import STILDumpCompilerInternalException
from .DomainUtils import DomainUtils
from .WFCUtils import WFCUtils
from .TimeUtils import TimeUtils
from .HashInfo import HashInfo
from .PattVecCmd import PattVecCmd
from .SigTimingInfo import SigTimingInfo
from .utils import get_line, get_col_error_pos



# -*- coding: utf-8 -*-
import inspect
from .DomainUtils import DomainUtils


class SelectorBlockParser:
    def __init__(self, debug=False):
        self.debug = debug

        self.curr_selector = DomainUtils.global_domain
        self.curr_var_name = None
        
        # Key is selector_name::variable_name
        # Value is Min/Typ/Max
        self.selector_var = {}

    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')

    def b_selector__SELECTOR_DOMAIN_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
            
        self.curr_selector = t.value

    def b_selector__OPEN_SELECTOR_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

    def b_selector__variable_name(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
            
        self.curr_var_name = str(t[0])

    def b_selector__variable_select(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
            
        key = self.curr_selector + "::" + self.curr_var_name
        
        self.selector_var[key] = str(t[0])

    def b_selector__CLOSE_SELECTOR_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
            
        self.curr_selector = DomainUtils.global_domain



# -*- coding: utf-8 -*-
import inspect


class ScanStructBlockParser:
    def __init__(self, debug=False):
        self.debug = debug

    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')

    def b_scan_structures__SCAN_STRUCTURES_DOMAIN_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

    def b_scan_structures__OPEN_SCAN_STRUCTURES_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

    def b_scan_structures__CLOSE_SCAN_STRUCTURES_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)



# -*- coding: utf-8 -*-


class HashInfo:

    POS_UNKNOWN = -1
    POS_BEFORE_SHIFT = 0
    POS_SHIFT = 1
    POS_AFTER_SHIFT = 2

    def __init__(self, pos_rel_shift, sig_ref, pos, debug=False):
        # Position relative to the shift
        self.pos_rel_shift = pos_rel_shift
        # Signal reference
        self.sig_ref = sig_ref
        # list of position of the # inside WFC list of the signal reference
        # [2, 5] if you have the following WFC list : 00#LX#X
        self.pos = pos
    
    def __str__(self):

        pos = "in unknwown position"
        if self.pos_rel_shift == HashInfo.POS_BEFORE_SHIFT:
            pos = "before the Shift block"
        elif self.pos_rel_shift == HashInfo.POS_AFTER_SHIFT:
            pos = "after the Shift block"
        elif self.pos_rel_shift == HashInfo.POS_SHIFT:
            pos = "in the Shift block"
            
        s = f"# at pos {self.pos} for signal/signal group {self.sig_ref} "+pos
        return s


# -*- coding: utf-8 -*-


class PattVecCmd:
    
    '''
    Contains pattern statements information for one vector address.
    On one vector address is possible to have multiple pattern statements.
    Like Label + Vector or WaveformTable + Vector ...
    The order of commands is not perserved. 
    '''

    CMD_NONE = 0
    CMD_LABEL = 1
    CMD_WFT = 2
    CMD_LOAD_LOOP_COUNTER = 3
    CMD_LOAD_MATCHLOOP_COUNTER = 4
    CMD_START_LOOP = 5
    CMD_START_MATCHLOOP = 6
    CMD_VECTOR = 7
    CMD_GOTO = 8
    CMD_BREAKPOINT = 9
    CMD_IDDQTESTPOINT = 10
    CMD_STOP = 11
    CMD_START_SHIFT = 12
    CMD_STOP_SHIFT = 13
    CMD_STOP_LOOP = 14
    CMD_STOP_COUNT_MATCHLOOP = 15
    CMD_STOP_INF_MATCHLOOP = 16
    CMD_MACRO = 17
    CMD_CALL = 18

    cmds = {
        0: "NO CMD",
        1: "LABEL",
        2: "WFT",
        3: "LOAD_LOOP_COUNTER",
        4: "LOAD_MATCHLOOP_COUNTER",
        5: "START_LOOP",
        6: "START_MATCHLOOP",
        7: "VECTOR",
        8: "GOTO",
        9: "BREAKPOINT",
        10: "IDDQTESTPOINT",
        11: "STOP",
        12: "START_SHIFT",
        13: "STOP_SHIFT",
        14: "STOP_LOOP",
        15: "STOP_COUNT_MATCHLOOP",
        16: "STOP_INF_MATCHLOOP",
        17: "MACRO",
        18: "CALL",
    }

    def __init__(self, debug=False):

        self.debug = debug
        self.cmd = 0
        # Key is the command number
        # Value is the command's value
        self.cmd_value = {}
        # Key is the command number
        # Value is dict with the command property's key and value
        self.cmd_prop = {}

    def add_cmd(self, cmd, value=None):
        self.cmd |= 1 << cmd
        if value is not None:
            self.set_value(cmd, value)

    def del_cmd(self, cmd):

        if cmd in self.cmd_value:
            self.cmd_value[cmd] = None
        self.cmd &= ~(1 << cmd)

    def get_cmd_name(cmd_id):
        return PattVecCmd.cmds[cmd_id]

    def have_cmd(self, cmd_id):
        found = False
        for bit in range(len(PattVecCmd.cmds)):
            if ((self.cmd >> bit) & 0x1) == 0x1:
                if cmd_id == bit:
                    found = True
                    break
        return found

    def get_cmd_ids(self):
        cmds = []
        for bit in range(len(PattVecCmd.cmds)):
            if ((self.cmd >> bit) & 0x1) == 0x1:
                cmds.append(bit)
        return cmds

    def set_value(self, cmd, value):
        self.cmd_value[cmd] = value

    def get_value(self, cmd_id):
        if cmd_id in self.cmd_value:
            return self.cmd_value[cmd_id]

    def add_prop(self, cmd, prop_name, prop_value):
        #print(f"add prop for cmd {PattVecCmd.get_cmd_name(cmd)}")
        if cmd in self.cmd_prop:
            d = self.cmd_prop[cmd]
            d[prop_name] = prop_value
            #print(f"{d}")
        else:
            #print("add new cmd for prop")
            self.cmd_prop[cmd] = {}
            d = self.cmd_prop[cmd]
            d[prop_name] = prop_value
            #print(f"{d}")

    def get_props(self, cmd_id):
        #cmd_name = PattVecCmd.get_cmd_name(cmd_id)
        #print(f"get_props for cmd {cmd_name}")
        if cmd_id in self.cmd_prop:
            return self.cmd_prop[cmd_id]
        
    def __str__(self):
        msg = ""
        cmds = self.get_cmd_ids()
        for cmd in cmds:
            value = self.get_value(cmd)
            cmd_name = PattVecCmd.get_cmd_name(cmd)
            
            if value == None:
                msg += f"{cmd_name}"
            else:
                msg += f"{cmd_name}={value} "
            msg += "|"
        return msg



# -*- coding: utf-8 -*-
import inspect

from .DomainUtils import DomainUtils
from .TimeUtils import TimeUtils

class PatternExecBlockParser:
    def __init__(self, debug=False):
        self.debug = debug
        
        self.op = {'+', '-', '/', '*', '(', ')'}

        # dict key   is the pattern exec block name
        # dict value is a list with pattern burst blocks
        self.patt_exec_block2patt_burst = {}

        self.curr_patt_exec = DomainUtils.global_domain

        self.curr_timing_domain = DomainUtils.global_domain

        # List with names of all pattern block which should be executed
        self.used_patterns = []

        # dict key   is the pattern exec block name
        # dict value is the timing domain
        self.patt_exec_block2time_domain = {}

        # dict key   is the pattern exec block name
        # dict value is the spec category
        self.patt_exec2category = {}

        # dict key   is the pattern exec block name
        # dict value is the spec selector
        self.patt_exec2selector = {}

        self.is_patternexec_block_defined = False

    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')

    def b_pattern_exec__PATTERN_EXEC_BLOCK_NAME(self, t):

        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if t in self.patt_exec_block2patt_burst:
            err_msg = f"Pattern exec block {t} already defined! Pattern exec block name must be unique."
            raise Exception(err_msg)
        else:
            self.patt_exec_block2patt_burst[t.value] = []
            self.curr_patt_exec = t.value

    def b_pattern_exec__OPEN_PATTERN_EXEC_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if self.is_signal_block_defined == False:
            err_msg = "Expected Signal block before any PatternExec block"
            raise Exception(err_msg)
        if self.is_signalgroups_block_defined == False:
            err_msg = "Expected SignalGroups block before any PatternExec block"
            raise Exception(err_msg)
        if self.is_timing_block_defined == False:
            err_msg = "Expected Timing block before any PatternExec block"
            raise Exception(err_msg)
        if self.is_patternburst_block_defined == False:
            err_msg = "Expected PatternBurst block before any PatternExec block"
            raise Exception(err_msg)
        self.is_patternexec_block_defined = True

        global_domain = DomainUtils.global_domain

        if global_domain in self.patt_exec_block2patt_burst:
            err_msg = "Only one global PatternExec block can exists!"
            raise Exception(err_msg)

        if self.curr_patt_exec == global_domain:
            self.patt_exec_block2patt_burst[self.curr_patt_exec] = []

    def b_pattern_exec__PATTERN_BURST_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        patt_burst_name = t.value

        if patt_burst_name not in self.patt_burst_block2pattern_blocks:
            err_msg = f"Pattern burst block named {patt_burst_name} is not defined!"
            raise Exception(err_msg)

        burst_names = self.patt_exec_block2patt_burst[self.curr_patt_exec]
        burst_names.append(patt_burst_name)

        patterns = self.patt_burst_block2pattern_blocks.get(patt_burst_name)
        for pattern in patterns:

            if pattern not in self.used_patterns:
                self.used_patterns.append(pattern)

            if pattern not in self.patt2timing_domain:
                self.patt2timing_domain[pattern] = [self.curr_timing_domain]
            else:
                domains = self.patt2timing_domain[pattern]
                if self.curr_timing_domain not in domains:
                    if domains is None:
                        self.patt2timing_domain[pattern] = [self.curr_timing_domain]
                    else:
                        domains.append(self.curr_timing_domain)

        macro_domain = self.patt_burst2macro_domain[patt_burst_name]
        if macro_domain in self.macro_domain2timing_domain:
            l = self.macro_domain2timing_domain[macro_domain]
            if self.curr_timing_domain not in l:
                l.append(self.curr_timing_domain)
        else:
            self.macro_domain2timing_domain[macro_domain] = [self.curr_timing_domain]

        proc_domain = self.patt_burst2proc_domain[patt_burst_name]
        if proc_domain in self.proc_domain2timing_domain:
            l = self.proc_domain2timing_domain[proc_domain]
            if self.curr_timing_domain not in l:
                l.append(self.curr_timing_domain)
        else:
            self.proc_domain2timing_domain[proc_domain] = [self.curr_timing_domain]

    def b_pattern_exec__TIMING_DOMAIN(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_timing_domain = t.value

        if self.curr_timing_domain not in self.time_domain2wft:
            err_msg = f"Timing domain {self.curr_timing_domain} is not defined!"
            raise Exception(err_msg)
            
    def b_pattern_exec__category_name(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
           
        self.patt_exec2category[self.curr_patt_exec] = str(t[0])
            
    def b_pattern_exec__selector_name(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.patt_exec2selector[self.curr_patt_exec] = str(t[0])

    def b_pattern_exec__CLOSE_PATTERN_EXEC_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.process_timings()

        self.patt_exec_block2time_domain[self.curr_patt_exec] = self.curr_timing_domain

        self.curr_patt_exec = DomainUtils.global_domain
        self.curr_timing_domain = DomainUtils.global_domain      
        

    def process_timings(self):
        
        for time_wft in self.wft2period:
            period = self.wft2period[time_wft]
#            key = self.curr_patt_exec + "::" + time_wft
#            print(f"key {key} period {period}")
            self.wft2period[time_wft] = self.parse_time_expr(period)
            
        
        # dict key   is the timing_domain::wft::signal_name
        # dict value is a SigTimingInfo object with timing information 
        for tws in self.sig_wft2timing:
            sti = self.sig_wft2timing[tws]
            wfcs = sti.get_wfcs()
            for wfc in wfcs:
                timing = sti.get_timing_for_wfc(wfc)
                for wfe_time in timing:
                    wfe = wfe_time[0]
                    time = wfe_time[1]
                    ftime = self.parse_time_expr(time)
                    sti.replace_timing_for_wfc(wfc, wfe, time, ftime)
        
            
    def get_var_value(self, selector, category, variable):
        
        sel_var = selector + "::" + variable
#        print(f"self.selector_var {self.selector_var}")
        if sel_var in self.selector_var:
            sel = self.selector_var[sel_var]
            # key is category::variable      
            if sel == 'Min':
                key = category + "::" + variable
                keyn = "NONE::" + variable
                if key in self.var_min_value:
                    return self.var_min_value[key]
                elif keyn in self.var_min_value:
                    return self.var_min_value[keyn]
                else:
                    err_msg = f"ERROR: Can not find 'Min' value for variable '{variable}' in category '{category}'!"
                    raise Exception(err_msg)
            elif sel == 'Typ':
                key = category + "::" + variable
                keyn = "NONE::" + variable
                if key in self.var_typ_value:
                    return self.var_typ_value[key]
                elif keyn in self.var_typ_value:
                    return self.var_typ_value[keyn]
                else:
                    err_msg = f"ERROR: Can not find 'Typ' value for variable '{variable}' in category '{category}'!"
                    raise Exception(err_msg)
            elif sel == 'Max':
                key = category + "::" + variable
                keyn = "NONE::" + variable
                if key in self.var_max_value:
                    return self.var_max_value[key]
                elif keyn in self.var_max_value:
                    return self.var_min_value[keyn]
                else:
                    err_msg = f"ERROR: Can not find 'Max' value for variable '{variable}' in category '{category}'!"
                    raise Exception(err_msg)
            else:
                err_msg = f"ERROR: Unknown selector {sel}!"
                raise Exception(err_msg)

                
        else:
            err_msg = f"ERROR: Variable {variable} is not defined in the selector {selector}!"
            raise Exception(err_msg)
            
            
    def parse_time_expr(self, time_expr):
        
        # Check first if the value is simple time:
        is_simple_time = True
        for i in range(0, len(time_expr)):
            char = time_expr[i]
            if char in self.op:
                is_simple_time = False
                break
        
        if is_simple_time:
            fsec = TimeUtils.get_time_fsec(time_expr)
            try:
                value = int(fsec)
                return str(value) + "fs"
            except Exception:
                # The value is spec variable, will be processed in the next lines
                pass

        # The time is expressoin, let's calculate it:
        cat = self.patt_exec2category[self.curr_patt_exec]
        sel = self.patt_exec2selector[self.curr_patt_exec]
            
        op_list = []

        buff = ""
        for i in range(0, len(time_expr)):
            char = time_expr[i]
            if char.isalnum():
                buff += char
            elif char == ' ':
                pass
            elif char == '_':
                buff += char
            elif char == '.':
                buff += char
            elif char in self.op:
                if buff in self.variables:
                    val = self.get_var_value(sel, cat, buff)
                    buff = val
                if len(buff) > 0:
                    op_list.append(buff)
                if len(char) > 0:
                    op_list.append(char)
                buff = ''
                
        if buff in self.variables:
            val = self.get_var_value(sel, cat, buff)
            if len(val) > 0:
                op_list.append(val)
        else:
            if len(buff) > 0:
                op_list.append(buff)
            
        # Return expression according order of operations
        return TimeUtils.bodmas(op_list, time_expr)

